from enum import Enum
from typing import Optional

from pydantic import SerializeAsAny

from ..base import NBCXBaseModel


class NBCXLayoutHorizontalAlignment(str, Enum):
    left = "left"
    center = "center"
    right = "right"


class NBCXLayoutVerticalAlignment(str, Enum):
    top = "top"
    center = "center"
    bottom = "bottom"


class NBCXLayoutFontWeight(str, Enum):
    normal = "normal"
    bold = "bold"


class NBCXLayoutJustify(str, Enum):
    normal = "normal"
    center = "center"
    start = "start"
    end = "end"
    left = "left"
    right = "right"
    space_between = "space-between"
    space_around = "space-around"
    space_evenly = "space-evenly"


class NBCXLayoutFontStyle(str, Enum):
    normal = "normal"
    italic = "italic"


class NBCXLayoutTextDecoration(str, Enum):
    none = "none"
    underline = "underline"


class NBCXTextComponent(NBCXBaseModel):
    content: Optional[str] = ""
    horizontal_alignment: Optional[SerializeAsAny[NBCXLayoutHorizontalAlignment]] = None
    vertical_alignment: Optional[SerializeAsAny[NBCXLayoutVerticalAlignment]] = None
    font_weight: Optional[SerializeAsAny[NBCXLayoutFontWeight]] = None
    font_style: Optional[SerializeAsAny[NBCXLayoutFontStyle]] = None
    text_decoration: Optional[SerializeAsAny[NBCXLayoutTextDecoration]] = None



from typing import List, Union

from nbformat import NotebookNode


def _append_or_extend(cells: list, cell_or_cells: Union[NotebookNode, List[NotebookNode]]) -> None:
    if isinstance(cell_or_cells, list):
        cells.extend(cell_or_cells)
    elif cell_or_cells:
        cells.append(cell_or_cells)
    # None, ignore



import os

from ipython_genutils.py3compat import getcwd
from nbconvert.exporters import PDFExporter
from nbconvert.exporters.pdf import LatexFailed
from nbconvert.filters.highlight import Highlight2Latex
from testpath.tempdir import TemporaryWorkingDirectory

from .template import TemplateOverrideMixin


class NBCXPDFExporter(TemplateOverrideMixin, PDFExporter):
    """Custom Exporter to reformat the notebook node to allow for document-level configuration
    from cell outputs based on tags"""

    export_from_notebook = "NBCX PDF Report via LaTeX"

    def from_notebook_node(self, nb, resources=None, **kw):
        # ********************************************** #
        # From LatexExporter
        # https://github.com/jupyter/nbconvert/blob/master/nbconvert/exporters/latex.py
        langinfo = nb.metadata.get("language_info", {})
        lexer = langinfo.get("pygments_lexer", langinfo.get("name", None))
        highlight_code = self.filters.get("highlight_code", Highlight2Latex(pygments_lexer=lexer, parent=self))
        self.register_filter("highlight_code", highlight_code)
        # ********************************************** #

        # ***************** CUSTOM CODE **************** #
        # call the overridden from_notebook_node
        latex, resources = super()._from_notebook_node_override(nb, resources=resources, **kw)
        # ********************************************** #

        # ********************************************** #
        # From PDFExporter
        # set texinputs directory, so that local files will be found
        if resources and resources.get("metadata", {}).get("path"):
            self.texinputs = resources["metadata"]["path"]
        else:
            self.texinputs = getcwd()

        self._captured_outputs = []
        with TemporaryWorkingDirectory():
            notebook_name = "notebook"
            resources["output_extension"] = ".tex"
            tex_file = self.writer.write(latex, resources, notebook_name=notebook_name)
            self.log.info("Building PDF")
            self.run_latex(tex_file)
            if self.run_bib(tex_file):
                self.run_latex(tex_file)

            pdf_file = notebook_name + ".pdf"
            if not os.path.isfile(pdf_file):
                raise LatexFailed("\n".join(self._captured_output))
            self.log.info("PDF successfully created")
            with open(pdf_file, "rb") as f:
                pdf_data = f.read()

        # convert output extension to pdf
        # the writer above required it to be tex
        resources["output_extension"] = ".pdf"
        # clear figure outputs, extracted by latex export,
        # so we don't claim to be a multi-file export.
        resources.pop("outputs", None)

        return pdf_data, resources
        # ********************************************** #



from typing import TYPE_CHECKING, List, Optional, Union

from nbformat import NotebookNode
from pydantic import SerializeAsAny, validator

from ..base import NBCXBaseModel, NBCXType
from ..layout import NBCXLayout

if TYPE_CHECKING:
    from ..config import NBCXConfiguration


class NBCXContent(NBCXBaseModel):
    content: Optional[Union[str, List[SerializeAsAny[NBCXBaseModel]]]] = ""
    layout: Optional[SerializeAsAny[NBCXLayout]] = None

    def generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["NBCXConfiguration"] = None,
        parent: Optional["NBCXBaseModel"] = None,
        attr: str = "",
        counter: Optional[int] = None,
    ) -> Optional[Union[NotebookNode, List[NotebookNode]]]:
        if isinstance(self.content, list):
            cells = [
                c.generate(metadata=metadata, config=config, parent=self, attr="content", counter=i)
                for i, c in enumerate(self.content)
            ]
        else:
            cell = super()._base_generate(
                metadata=metadata, config=config, parent=parent, attr="content", counter=counter
            )
            cell.metadata.tags.append("nbcx:content")
            cell.metadata.nbcx.role = "content"
            cell.metadata.nbcx.data = ""
            cell.source = self.content
            cells = [cell]
        return cells

    @validator("content", pre=True)
    def convert_content_from_obj(cls, v):
        if v is None:
            return []
        if isinstance(v, list):
            for i, element in enumerate(v):
                if isinstance(element, str):
                    v[i] = NBCXContent(type=NBCXType.from_string(element))
                elif isinstance(element, dict):
                    v[i] = NBCXBaseModel._to_type(element)
        return v


class NBCXContentMarkdown(NBCXContent):
    content: Optional[str] = ""

    def generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["NBCXConfiguration"] = None,
        parent: Optional["NBCXBaseModel"] = None,
        attr: str = "",
        counter: Optional[int] = None,
    ) -> Optional[Union[NotebookNode, List[NotebookNode]]]:
        cell = super()._base_generate_md(metadata=metadata)
        cell.metadata.tags.append("nbcx:content")
        cell.metadata.nbcx.role = "content"
        cell.source = self.content
        return cell


class NBCXContentCode(NBCXContent):
    ...


class NBCXContentDynamic(NBCXContent):
    content: Optional[str] = ""

    def generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["NBCXConfiguration"] = None,
        parent: Optional["NBCXBaseModel"] = None,
        attr: str = "",
        counter: Optional[int] = None,
    ) -> Optional[Union[NotebookNode, List[NotebookNode]]]:
        # force parent to None to load from json explicitly
        cell = super()._base_generate(
            metadata=metadata,
            config=config,
            parent=parent,
            attr=attr,
            counter=counter,
            call_with_context=config.context.nb_var_name,
        )

        # add tags and role
        cell.metadata.tags.extend(["nbcx:content", "nbcx:content:dynamic"])
        cell.metadata.nbcx.role = "content"
        return cell



import os

from nbconvert.exporters import TemplateExporter


class TemplateOverrideMixin:
    def _from_notebook_node_override(self, nb, resources=None, **kw):
        # make sure this is set prior to execution
        os.environ["NBCX_NBCONVERT"] = "yes"

        # ********************************************** #
        # From Template Exporter
        # https://github.com/jupyter/nbconvert/blob/master/nbconvert/exporters/templateexporter.py
        # Call Exporter's from_notebook_node
        nb_copy, resources = super(TemplateExporter, self).from_notebook_node(nb, resources, **kw)
        resources.setdefault("raw_mimetypes", self.raw_mimetypes)
        resources["global_content_filter"] = {
            "include_code": not self.exclude_code_cell,
            "include_markdown": not self.exclude_markdown,
            "include_raw": not self.exclude_raw,
            "include_unknown": not self.exclude_unknown,
            "include_input": not self.exclude_input,
            "include_output": not self.exclude_output,
            "include_input_prompt": not self.exclude_input_prompt,
            "include_output_prompt": not self.exclude_output_prompt,
            "no_prompt": self.exclude_input_prompt and self.exclude_output_prompt,
        }
        # ********************************************** #

        # ***************** CUSTOM CODE **************** #
        # Inject our own vars
        resources["nbcx"] = {}
        resources["nbcx"]["headers"] = {}

        for i, cell in enumerate(nb_copy.cells):
            tags = [t for t in cell.metadata.get("tags", []) if t.startswith("nbcx_")]
            if len(tags) > 1:
                raise Exception("Can't mix nbcx tags!")

            if not tags:
                # don't do anything
                continue

            tag = tags[0]
            if tag in ("nbcx_title", "nbcx_ignore"):
                # don't do anything with these
                continue
            elif tag in ("nbcx_title",):
                resources["nbcx"][tag.replace("nbcx_", "")] = cell

            elif tag in (
                "nbcx_lhead",
                "nbcx_chead",
                "nbcx_rhead",
                "nbcx_lfoot",
                "nbcx_cfoot",
                "nbcx_rfoot",
            ):
                if tag in ("nbcx_lhead", "nbcx_chead", "nbcx_rhead"):
                    resources["nbcx"]["headerrule"] = True
                if tag in ("nbcx_lfoot", "nbcx_cfoot", "nbcx_rfoot"):
                    resources["nbcx"]["footerrule"] = True

                resources["nbcx"]["headers"][tag.replace("nbcx_", "")] = cell
        # ********************************************** #

        # ********************************************** #
        # Top level variables are passed to the template_exporter here.
        output = self.template.render(nb=nb_copy, resources=resources)
        output = output.lstrip("\r\n")
        return output, resources
        # ********************************************** #



from nbconvert.exporters import HTMLExporter
from nbconvert.filters.highlight import Highlight2HTML

from .template import TemplateOverrideMixin


class NBCXHTMLExporter(TemplateOverrideMixin, HTMLExporter):
    export_from_notebook = "NBCX HTML"

    def from_notebook_node(self, nb, resources=None, **kw):
        # ********************************************** #
        # From HTMLExporter
        # https://github.com/jupyter/nbconvert/blob/master/nbconvert/exporters/html.py
        langinfo = nb.metadata.get("language_info", {})
        lexer = langinfo.get("pygments_lexer", langinfo.get("name", None))
        highlight_code = self.filters.get("highlight_code", Highlight2HTML(pygments_lexer=lexer, parent=self))
        self.register_filter("highlight_code", highlight_code)
        # ********************************************** #

        # ***************** CUSTOM CODE **************** #
        return super()._from_notebook_node_override(nb, resources, **kw)
        # ********************************************** #



from IPython.display import Latex, display
from nbconvert.filters.pandoc import convert_pandoc


def latex(text, color="", **kwargs):
    """print in latex"""
    if color:
        text = "\\textcolor{%s}{%s}" % (color, text)
    return Latex(text + "\n", **kwargs)


def print(text, **kwargs):
    """wrapper around printing"""
    if not isinstance(text, str):
        display(text)
        return
    display(
        latex(
            convert_pandoc(text + "\\newline", "markdown+tex_math_double_backslash", "latex"),
            **kwargs,
        )
    )


def hr():
    """horizontal rule"""
    return latex("\\noindent\\makebox[\\linewidth]{\\rule{\\paperwidth - 1cm}{0.4pt}}")


def newpage():
    """make a new page. in html, this just does a horizontal rule"""
    return latex("\\newpage")


def table(df, title="", footnote=""):
    """helper to display a table"""
    return latex(
        "\\begin{center} "
        "\\begin{threeparttable}"
        "\\caption{" + title + "}" + df.to_latex(escape=False) + "\\begin{tablenotes}"
        "\\small"
        "\\item " + footnote + "\\end{tablenotes}"
        "\\end{threeparttable}"
        "\\end{center}"
    )


def pagenum():
    """display a page number (latex only)"""
    return latex("\\thepage")


def _make(text, h_type):
    return convert_pandoc("#" * h_type + " " + text + "\n", "markdown+tex_math_double_backslash", "latex")


def p(text, **kwargs):
    return Latex(text, **kwargs)


def h1(text, **kwargs):
    return Latex(_make(text, 1), **kwargs)


def h2(text, **kwargs):
    return Latex(_make(text, 2), **kwargs)


def h3(text, **kwargs):
    return Latex(_make(text, 3), **kwargs)


def h4(text, **kwargs):
    return Latex(_make(text, 4), **kwargs)


def h5(text, **kwargs):
    return Latex(_make(text, 5), **kwargs)


def grid(items_and_weights):
    print(Latex("\n\\begin{Row}%\n"))
    for val, width in items_and_weights:
        print(Latex("\n\\begin{Cell}{" + str(width) + "}\n"))

        if not isinstance(val, str):
            print(Latex("\\vspace*{-.4cm}\n"))
        print(val)
        print(Latex("\n\n\\end{Cell}\n"))
    print(Latex("\n\\end{Row}\n"))



"""
BSD 3-Clause License

Copyright (c) 2020, Tim Head <betatim@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import asyncio
import concurrent.futures
import os
import tempfile

import nbconvert
from nbconvert.exporters import Exporter
from pyppeteer import launch
from traitlets import default


async def html_to_pdf(html_file, pdf_file):
    """Convert a HTML file to a PDF"""
    browser = await launch(handleSIGINT=False, handleSIGTERM=False, handleSIGHUP=False)
    page = await browser.newPage()
    await page.setViewport(dict(width=994, height=768))
    await page.emulateMedia("screen")

    await page.goto(f"file:///{html_file}", {"waitUntil": ["networkidle2"]})

    page_margins = {
        "left": "0px",
        "right": "0px",
        "top": "0px",
        "bottom": "0px",
    }

    dimensions = await page.evaluate(
        """() => {
        return {
            width: document.body.scrollWidth,
            height: document.body.scrollHeight,
            offsetHeight: document.body.offsetHeight,
            deviceScaleFactor: window.devicePixelRatio,
        }
    }"""
    )
    width = dimensions["width"]
    height = dimensions["height"]

    await page.addStyleTag(
        {
            "content": """
                #notebook-container {
                    box-shadow: none;
                    padding: unset
                }
                div.cell {
                    page-break-inside: avoid;
                }
                div.output_wrapper {
                    page-break-inside: avoid;
                }
                div.output {
                    page-break-inside: avoid;
                }
         """
        }
    )

    await page.pdf(
        {
            "path": pdf_file,
            "width": width,
            # Adobe can not display pages longer than 200inches. So we limit
            # ourselves to that and start a new page if needed.
            "height": min(height, 200 * 72),
            "printBackground": True,
            "margin": page_margins,
        }
    )

    await browser.close()


async def notebook_to_pdf(notebook, pdf_path, config=None, resources=None, **kwargs):
    """Convert a notebook to PDF"""
    if config is None:
        config = {}
    exporter = nbconvert.HTMLExporter(config=config)
    exported_html, _ = exporter.from_notebook_node(notebook, resources=resources, **kwargs)

    with tempfile.NamedTemporaryFile(suffix=".html") as f:
        f.write(exported_html.encode())
        f.flush()
        await html_to_pdf(f.name, pdf_path)


class PDFExporter(Exporter):
    """Convert a notebook to a PDF

    Expose this package's functionality to nbconvert
    """

    # a thread pool to run our async event loop. We use our own
    # because `from_notebook_node` isn't async but sometimes is called
    # inside a tornado app that already has an event loop
    pool = concurrent.futures.ThreadPoolExecutor()

    export_from_notebook = "PDF via HTML"
    output_mimetype = "application/pdf"

    @default("file_extension")
    def _file_extension_default(self):
        return ".pdf"

    def __init__(self, config=None, **kw):
        with_default_config = self.default_config
        if config:
            with_default_config.merge(config)

        super().__init__(config=with_default_config, **kw)

    def from_notebook_node(self, notebook, resources=None, **kwargs):
        notebook, resources = super().from_notebook_node(notebook, resources=resources, **kwargs)

        # if it is unset or an empty value, set it
        if resources.get("ipywidgets_base_url", "") == "":
            resources["ipywidgets_base_url"] = "https://unpkg.com/"

        with tempfile.TemporaryDirectory(suffix="nb-as-pdf") as name:
            pdf_fname = os.path.join(name, "output.pdf")

            self.pool.submit(
                asyncio.run,
                notebook_to_pdf(
                    notebook,
                    pdf_fname,
                    config=self.config,
                    resources=resources,
                    **kwargs,
                ),
            ).result()
            resources["output_extension"] = ".pdf"

            with open(pdf_fname, "rb") as f:
                pdf_bytes = f.read()

        return (pdf_bytes, resources)



from typing import TYPE_CHECKING, List, Optional, Union

from nbformat import NotebookNode
from pydantic import SerializeAsAny, validator

from ..base import NBCXBaseModel
from ..utils import _append_or_extend
from .common import NBCXLayoutHorizontalAlignment, NBCXLayoutJustify, NBCXLayoutVerticalAlignment
from .page_number import NBCXLayoutPageNumber

if TYPE_CHECKING:
    from ..config import NBCXConfiguration


class NBCXLayout(NBCXBaseModel):
    vertical_alignment: Optional[SerializeAsAny[NBCXLayoutVerticalAlignment]] = None
    horizontal_alignment: Optional[SerializeAsAny[NBCXLayoutHorizontalAlignment]] = None
    justify: Optional[SerializeAsAny[NBCXLayoutJustify]] = None


class NBCXLayoutRegionBase(NBCXLayout):
    page_number: Optional[SerializeAsAny[NBCXLayoutPageNumber]] = None

    def _generate_helper(
        self,
        attr,
        metadata: Optional[dict] = None,
        config: Optional["NBCXConfiguration"] = None,
        parent: Optional["NBCXBaseModel"] = None,
    ) -> NotebookNode:
        cell = super()._base_generate(metadata=metadata, config=config, parent=parent, attr=attr)
        cell.metadata.tags.extend(["nbcx:layout", f"nbcx:layout:{attr}"])
        cell.metadata.nbcx.role = "layout"
        cell.metadata.nbcx.ignore = True

        return cell


class NBCXLayoutHeader(NBCXLayoutRegionBase):
    def generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["NBCXConfiguration"] = None,
        parent: Optional["NBCXBaseModel"] = None,
    ) -> NotebookNode:
        return self._generate_helper("header", metadata=metadata, config=config, parent=parent)


class NBCXLayoutFooter(NBCXLayoutRegionBase):
    def generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["NBCXConfiguration"] = None,
        parent: Optional["NBCXBaseModel"] = None,
    ) -> NotebookNode:
        return self._generate_helper("footer", metadata=metadata, config=config, parent=parent)


class NBCXLayoutLeft(NBCXLayoutRegionBase):
    def generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["NBCXConfiguration"] = None,
        parent: Optional["NBCXBaseModel"] = None,
    ) -> NotebookNode:
        return self._generate_helper("left", metadata=metadata, config=config, parent=parent)


class NBCXLayoutRight(NBCXLayoutRegionBase):
    def generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["NBCXConfiguration"] = None,
        parent: Optional["NBCXBaseModel"] = None,
    ) -> NotebookNode:
        return self._generate_helper("right", metadata=metadata, config=config, parent=parent)


class NBCXLayoutTopLeft(NBCXLayoutRegionBase):
    def generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["NBCXConfiguration"] = None,
        parent: Optional["NBCXBaseModel"] = None,
    ) -> NotebookNode:
        return self._generate_helper("top_left", metadata=metadata, config=config, parent=parent)


class NBCXLayoutTopRight(NBCXLayoutRegionBase):
    def generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["NBCXConfiguration"] = None,
        parent: Optional["NBCXBaseModel"] = None,
    ) -> NotebookNode:
        return self._generate_helper("top_right", metadata=metadata, config=config, parent=parent)


class NBCXLayoutBottomLeft(NBCXLayoutRegionBase):
    def generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["NBCXConfiguration"] = None,
        parent: Optional["NBCXBaseModel"] = None,
    ) -> NotebookNode:
        return self._generate_helper("bottom_left", metadata=metadata, config=config, parent=parent)


class NBCXLayoutBottomRight(NBCXLayoutRegionBase):
    def generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["NBCXConfiguration"] = None,
        parent: Optional["NBCXBaseModel"] = None,
    ) -> NotebookNode:
        return self._generate_helper("bottom_right", metadata=metadata, config=config, parent=parent)


class NBCXLayoutGlobal(NBCXBaseModel):
    header: Optional[SerializeAsAny[NBCXLayoutHeader]] = None
    footer: Optional[SerializeAsAny[NBCXLayoutFooter]] = None
    left: Optional[SerializeAsAny[NBCXLayoutLeft]] = None
    right: Optional[SerializeAsAny[NBCXLayoutRight]] = None
    top_left: Optional[SerializeAsAny[NBCXLayoutTopLeft]] = None
    top_right: Optional[SerializeAsAny[NBCXLayoutTopRight]] = None
    bottom_left: Optional[SerializeAsAny[NBCXLayoutBottomLeft]] = None
    bottom_right: Optional[SerializeAsAny[NBCXLayoutBottomRight]] = None

    @validator("header", pre=True)
    def convert_header_from_obj(cls, v):
        return NBCXBaseModel._to_type(v, NBCXLayoutHeader)

    @validator("footer", pre=True)
    def convert_footer_from_obj(cls, v):
        return NBCXBaseModel._to_type(v, NBCXLayoutFooter)

    @validator("left", pre=True)
    def convert_left_from_obj(cls, v):
        return NBCXBaseModel._to_type(v, NBCXLayoutLeft)

    @validator("right", pre=True)
    def convert_right_from_obj(cls, v):
        return NBCXBaseModel._to_type(v, NBCXLayoutRight)

    @validator("top_left", pre=True)
    def convert_top_left_from_obj(cls, v):
        return NBCXBaseModel._to_type(v, NBCXLayoutTopLeft)

    @validator("top_right", pre=True)
    def convert_top_right_from_obj(cls, v):
        return NBCXBaseModel._to_type(v, NBCXLayoutTopRight)

    @validator("bottom_left", pre=True)
    def convert_bottom_left_from_obj(cls, v):
        return NBCXBaseModel._to_type(v, NBCXLayoutBottomLeft)

    @validator("bottom_right", pre=True)
    def convert_bottom_right_from_obj(cls, v):
        return NBCXBaseModel._to_type(v, NBCXLayoutBottomRight)

    def generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["NBCXConfiguration"] = None,
        parent: Optional["NBCXBaseModel"] = None,
    ) -> Optional[Union[NotebookNode, List[NotebookNode]]]:
        cells = []

        # parent and config should be equal for the global layout
        assert parent == config

        main_cell = super()._base_generate(metadata=metadata, config=config, parent=config, attr="layout")
        main_cell.metadata.tags.append("nbcx:layout")
        main_cell.metadata.nbcx.role = "layout"
        main_cell.metadata.nbcx.ignore = True
        cells.append(main_cell)

        for attr in (
            "header",
            "footer",
            "left",
            "right",
            "top_left",
            "top_right",
            "bottom_left",
            "bottom_right",
        ):
            if getattr(self, attr) is not None:
                # pass in `self` as `parent here`
                _append_or_extend(cells, getattr(self, attr).generate(metadata=metadata, config=config, parent=self))
        return cells



from typing import Optional

from IPython.display import HTML, display

from nbcx import NBCXContentCover, NBCXContentImage


class ExampleCoverPageContent(NBCXContentCover):
    logo: Optional[NBCXContentImage]
    title: Optional[str] = ""
    subtitle: Optional[str] = ""

    def __call__(self, ctx=None, *args, **kwargs):
        self.logo(ctx=ctx, *args, **kwargs)
        display(HTML(f"<h1>{self.title}</h1>"))
        display(HTML(f"<h2>{self.subtitle}</h2>"))
        display(HTML('<p class="pagebreak"></p>'))



from IPython.display import Image


def image(data=None, path=None, **kwargs):
    """display a image"""
    # measure in pixels for html, but cm for latex
    width = kwargs.get("width", "")
    height = kwargs.get("height", "")
    align = kwargs.pop("align", "left")
    metadata = kwargs.pop("metadata", {})

    if width:
        if isinstance(width, str) and not width.endswith("cm"):
            widthcm = width + "cm"
        elif not isinstance(width, str):
            # assume in pixels, get cm

            # TODO assume 96 DPI
            widthcm = "{}cm".format(int(width / 36))
        else:
            # assume already in cm, get pixels
            width, widthcm = float(width.replace("cm", "")), width

            # TODO assume 96 DPI
            width = int(width * 36)
            kwargs["width"] = width
    else:
        widthcm = ""

    if height:
        if isinstance(height, str) and not height.endswith("cm"):
            heightcm = height + "cm"
        elif not isinstance(height, str):
            # assume in pixels, get cm

            # TODO assume 96 DPI
            heightcm = "{}cm".format(int(height / 36))

        else:
            # assume already in cm, get pixels
            height, heightcm = float(height.replace("cm", "")), height

            # TODO assume 96 DPI
            height = int(height * 36)
            kwargs["height"] = height
    else:
        heightcm = ""

    metadata["widthcm"] = widthcm
    metadata["heightcm"] = heightcm
    metadata["align"] = align

    if path:
        return Image(filename=path, metadata=metadata, **kwargs)
    return Image(data=data, metadata=metadata, **kwargs)



from pathlib import Path
from typing import Dict, List, Union

from nbformat import NotebookNode
from omegaconf import DictConfig, OmegaConf
from pydantic import Field, PrivateAttr, SerializeAsAny, validator

from .base import NBCXBaseModel, NBCXType
from .content import NBCXContent
from .context import NBCXContext
from .layout import NBCXLayoutGlobal
from .outputs import NBCXOutputs
from .parameters import NBCXParameters
from .utils import _append_or_extend


class NBCXConfiguration(NBCXBaseModel):
    resources: Dict[str, SerializeAsAny[NBCXBaseModel]] = Field(default_factory=dict)
    outputs: SerializeAsAny[NBCXOutputs]
    parameters: SerializeAsAny[NBCXParameters]
    layout: SerializeAsAny[NBCXLayoutGlobal]
    context: SerializeAsAny[NBCXContext]
    content: List[SerializeAsAny[NBCXContent]] = Field(default_factory=list)

    # internals
    _nb_var_name: str = PrivateAttr(default="nbcx_config")

    @validator("resources", pre=True)
    def convert_resources_from_obj(cls, value):
        if value is None:
            value = {}
        if isinstance(value, dict):
            for k, v in value.items():
                value[k] = NBCXBaseModel._to_type(v)
        return value

    @validator("outputs", pre=True)
    def convert_outputs_from_obj(cls, v):
        return NBCXBaseModel._to_type(v, NBCXOutputs)

    @validator("parameters", pre=True)
    def convert_parameters_from_obj(cls, v):
        return NBCXBaseModel._to_type(v, NBCXParameters)

    @validator("layout", pre=True)
    def convert_layout_from_obj(cls, v):
        return NBCXBaseModel._to_type(v, NBCXLayoutGlobal)

    @validator("context", pre=True)
    def convert_context_from_obj(cls, v):
        return NBCXBaseModel._to_type(v, NBCXContext)

    @validator("content", pre=True)
    def convert_content_from_obj(cls, v):
        if v is None:
            return []
        if isinstance(v, list):
            for i, element in enumerate(v):
                if isinstance(element, str):
                    v[i] = NBCXContent(type=NBCXType.from_string(element))
                elif isinstance(element, dict):
                    v[i] = NBCXBaseModel._to_type(element)
        return v

    def generate(self, metadata: dict = None, config: "NBCXConfiguration" = None) -> List[NotebookNode]:
        cells = []

        # start with parameters for papermill compat
        _append_or_extend(cells, self.parameters.generate(metadata=metadata, config=self))

        # now do the configuration itself
        _append_or_extend(cells, self._generate_self(metadata=metadata))

        # now do
        # context: SerializeAsAny[NBCXContext]
        _append_or_extend(cells, self.context.generate(metadata=metadata, config=self, parent=self))

        # resources: Dict[str, SerializeAsAny[NBCXBaseModel]] = Field(default_factory=dict)
        # cell.metadata.nbcx.resources = {k: v.json() for k, v in self.resources.items()}

        # outputs: SerializeAsAny[NBCXOutputs]

        # now setup the layout
        # layout: SerializeAsAny[NBCXLayout]
        _append_or_extend(cells, self.layout.generate(metadata=metadata, config=self, parent=self))

        # content: List[SerializeAsAny[NBCXContent]] = Field(default_factory=list)
        for i, content in enumerate(self.content):
            _append_or_extend(
                cells, content.generate(metadata=metadata, config=self, parent=self, attr="content", counter=i)
            )
        return cells

    def _generate_self(self, metadata: dict = None) -> NotebookNode:
        cell = super()._base_generate(metadata=metadata, config=None)
        cell.metadata.tags.append("nbcx:configuration")
        cell.metadata.nbcx.role = "configuration"
        cell.metadata.nbcx.ignore = True

        # omit the data
        cell.metadata.nbcx.data = ""

        # add resources
        # TODO do this or no?
        # cell.metadata.nbcx.resources = {k: v.json() for k, v in self.resources.items()}
        cell.metadata.nbcx.outputs = self.outputs.json()
        return cell

    def _generate_resources_cells(self, metadata: dict = None):
        cell = super()._base_generate(metadata=metadata, config=None)
        cell.metadata.tags.append("nbcx:resources")
        cell.metadata.nbcx.role = "resources"
        cell.metadata.nbcx.ignore = True

        # omit the data
        cell.metadata.nbcx.data = ""

        # add resources
        # mod = ast.Module(body=[], type_ignores=[])
        # for k, v in self.resources.items():
        #     # TODO
        #     ...
        return cell


def load(path_or_model: Union[str, Path, dict, NBCXConfiguration]) -> NBCXConfiguration:
    if isinstance(path_or_model, NBCXConfiguration):
        return path_or_model

    if isinstance(path_or_model, str) and (path_or_model.endswith(".yml") or path_or_model.endswith(".yaml")):
        path_or_model = Path(path_or_model).resolve()

    if isinstance(path_or_model, Path):
        path_or_model = OmegaConf.load(path_or_model)

    if isinstance(path_or_model, DictConfig):
        container = OmegaConf.to_container(path_or_model, resolve=True, throw_on_missing=True)
        return NBCXConfiguration(**container)

    raise TypeError(f"Path or model malformed: {path_or_model} {type(path_or_model)}")



from .base import (
    NBCXLayout,
    NBCXLayoutBottomLeft,
    NBCXLayoutBottomRight,
    NBCXLayoutFooter,
    NBCXLayoutGlobal,
    NBCXLayoutHeader,
    NBCXLayoutLeft,
    NBCXLayoutRegionBase,
    NBCXLayoutRight,
    NBCXLayoutTopLeft,
    NBCXLayoutTopRight,
)
from .common import (
    NBCXLayoutFontStyle,
    NBCXLayoutFontWeight,
    NBCXLayoutHorizontalAlignment,
    NBCXLayoutJustify,
    NBCXLayoutTextDecoration,
    NBCXLayoutVerticalAlignment,
    NBCXTextComponent,
)
from .page_number import NBCXLayoutPageNumber



from nbcx import NBCXContentDynamic, print


class ExampleMarkdownBlock(NBCXContentDynamic):
    def __call__(self, ctx=None, *args, **kwargs):
        print(
            """# Performance
Now some discussion of the performance, as well as some charts
Lorem ipsum dolor sit amet, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
## A Chart
A chart of some data
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
"""
        )



import ast
from json import dumps
from typing import TYPE_CHECKING

from nbformat import NotebookNode

from .base import NBCXBaseModel

if TYPE_CHECKING:
    from .config import NBCXConfiguration


class NBCXParameters(NBCXBaseModel):
    def generate(self, metadata: dict = None, config: "NBCXConfiguration" = None) -> NotebookNode:
        cell = self._base_generate_meta(metadata=metadata)
        cell.metadata.tags.extend(["parameters", "nbcx:parameters"])
        cell.metadata.nbcx.role = "parameters"
        cell.metadata.nbcx.ignore = True

        mod = ast.Module(body=[], type_ignores=[])
        for i, (k, v) in enumerate(self.dict().items()):
            if k in ("type",):
                continue
            mod.body.append(
                ast.Assign(
                    targets=[ast.Name(id=k, ctx=ast.Store())],
                    value=ast.parse(dumps(v), mode="eval").body,
                    lineno=i,
                )
            )
        source = ast.unparse(mod).replace('"', '\\"')
        cell.source = source
        return cell



from io import BytesIO

import dominate.tags as dt
import dominate.util as du
from IPython.display import HTML, display
from ipywidgets import Box, Output
from nbconvert.filters.pandoc import convert_pandoc

from .image import Image

_p = print


def _html(text, color=""):
    """print in html"""
    text = convert_pandoc(text, "markdown+tex_math_double_backslash", "html")

    if color:
        d = dt.div()
        d.attributes["style"] = "color: {};".format(color)
        d.appendChild(du.raw(text))
    else:
        d = du.raw(text)

    return HTML(str(d))


def print(text, **kwargs):
    """wrapper around printing"""
    if not isinstance(text, str):
        display(text)
        return
    display(_html(text, **kwargs))


def hr():
    """horizontal rule"""
    return HTML(str(dt.hr()))


def newpage():
    """make a new page. in html, this just does a horizontal rule"""
    p = dt.p()
    p.attributes["style"] = "page-break-before: always;"
    return _html(str(p))


def table(df, title="", footnote=""):
    """helper to display a table"""
    ret = ""
    if title:
        ret += "### {}\n".format(title)
    ret += df.to_html()
    if footnote:
        ret += "\n" + footnote + "\n"
    return _html(ret)


def pagenum():
    """display a page number (latex only)"""
    # TODO
    return "[pagenum]"


def _make(text, h_type, **kwargs):
    h = getattr(dt, h_type)(text)
    h.attributes.update(**kwargs)
    return h


def p(text, **kwargs):
    return HTML(str(_make(text, "p", **kwargs)))


def h1(text, **kwargs):
    return HTML(str(_make(text, "h1", **kwargs)))


def h2(text, **kwargs):
    return HTML(str(_make(text, "h2", **kwargs)))


def h3(text, **kwargs):
    return HTML(str(_make(text, "h3", **kwargs)))


def h4(text, **kwargs):
    return HTML(str(_make(text, "h4", **kwargs)))


def h5(text, **kwargs):
    return HTML(str(_make(text, "h5", **kwargs)))


def _grid(items_and_weights):
    d = dt.div()
    d.attributes["style"] = "display: flex; flex-direction: row;"

    for val, width in items_and_weights:
        if isinstance(val, Image):
            sd = dt.img()
            sd.set_attribute("src", "data:image/png;base64,{}".format(val._repr_png_()))
        else:
            raw_html = val._repr_html_()
            sd = dt.div(du.raw(raw_html))
        sd.attributes["style"] = "flex: {};".format(width)
        d.appendChild(sd)
    return HTML(str(d))


def grid(items_and_weights):
    children = []
    for val, width in items_and_weights:
        out = Output(layout={"flex": str(width)})
        children.append(out)
        with out:
            print(val)
    return Box(children, layout={"display": "flex", "flex-direction": "row"})


def plot(fig):
    imgdata = BytesIO()
    fig.savefig(imgdata)
    imgdata.seek(0)
    return Image(imgdata.read())



from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING

from nbformat import NotebookNode
from pydantic import DirectoryPath, SerializeAsAny, validator

from .base import NBCXBaseModel

if TYPE_CHECKING:
    from .config import NBCXConfiguration


class NBCXOutputNaming(str, Enum):
    uuid = "uuid"
    sha = "sha"


class NBCXOutputs(NBCXBaseModel):
    path_root: SerializeAsAny[DirectoryPath]
    naming: SerializeAsAny[NBCXOutputNaming]

    @validator("path_root", pre=True)
    def convert_str_to_path(cls, v):
        if isinstance(v, str):
            v = Path(v)
        if isinstance(v, Path):
            v.mkdir(parents=True, exist_ok=True)
            return v
        raise TypeError

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.path_root.mkdir(parents=True, exist_ok=True)

    def generate(self, metadata: dict = None, config: "NBCXConfiguration" = None) -> NotebookNode:
        cell = super()._base_generate(metadata=metadata, config=config, attr="outputs")
        cell.metadata.tags.append("nbcx:outputs")
        cell.metadata.nbcx.role = "outputs"
        return cell



from typing import TYPE_CHECKING, Optional

from IPython.display import HTML, display
from nbformat import NotebookNode

from ..base import NBCXBaseModel
from .base import NBCXContent

if TYPE_CHECKING:
    from ..config import NBCXConfiguration


class NBCXContentTableOfContents(NBCXContent):
    def generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["NBCXConfiguration"] = None,
        parent: Optional[NBCXBaseModel] = None,
        attr: str = "",
        counter: Optional[int] = None,
    ) -> NotebookNode:
        cell = super()._base_generate(
            metadata=metadata,
            config=config,
            parent=parent,
            attr=attr,
            counter=counter,
            call_with_context=config.context.nb_var_name,
        )
        # add tags and role
        cell.metadata.tags.extend(["nbcx:content", "nbcx:content:table-of-contents"])
        cell.metadata.nbcx.role = "content"
        return cell

    def __call__(self, ctx=None, *args, **kwargs):
        display(HTML('<div id="toc"></div>'))
        # display(HTML('<p class="pagebreak"></p>'))



from enum import Enum


class Tags(Enum):
    # Title Page
    TITLEPAGE = "nbcx_title"

    # Header
    LHEAD = "nbcx_lhead"
    CHEAD = "nbcx_chead"
    RHEAD = "nbcx_rhead"
    HEADER = "nbcx_chead"  # center by default

    # Footer
    LFOOT = "nbcx_lfoot"
    CFOOT = "nbcx_cfoot"
    RFOOT = "nbcx_rfoot"
    FOOTER = "nbcx_cfoot"  # center by default

    # Utilities
    NBCONVERT_IGNORE = "nbcx_ignore"
    PARAMETERS = "parameters"  # for Papermill



import os
from functools import lru_cache


@lru_cache(None)
def in_nbconvert():
    """helper function to check if running in nbcx context"""
    return True if (os.environ.get("NBCX_NBCONVERT", "") or os.environ.get("NBCX_CONTEXT", "")) else False


@lru_cache(None)
def nbconvert_context():
    """get context in which nbcx is running, either 'pdf' or 'html'"""
    if in_nbconvert():
        if os.environ.get("NBCX_CONTEXT", "") == "html":
            return "html"
        return "pdf"
    return ""



from sys import version_info

from nbformat import NotebookNode
from nbformat.v4 import new_notebook

import nbcx

from .config import NBCXConfiguration


def generate(config: NBCXConfiguration) -> NotebookNode:
    nb = new_notebook()
    nb.metadata.nbcx = {}
    nb.metadata.nbcx.version = nbcx.__version__
    nb.metadata.nbcx.tags = []
    nb.metadata.nbcx.language = f"python{version_info.major}.{version_info.minor}"

    base_meta = {
        "tags": [],
        "nbcx": {},
    }

    nb.cells = config.generate(metadata=base_meta, config=config)

    return nb



from .base import NBCXBaseModel, NBCXType
from .config import NBCXConfiguration, load
from .content import *
from .context import NBCXContext
from .layout import *
from .outputs import NBCXOutputNaming, NBCXOutputs
from .parameters import NBCXParameters



from nbcx import NBCXContentDynamic, table


class ExampleCodeBlock(NBCXContentDynamic):
    def __call__(self, ctx=None, *args, **kwargs):
        return table(
            ctx.df[["A", "B", "C", "D"]].head(20),
            "Sample Table",
            "This is where authors provide additional information about the data, including whatever notes are needed.",
        )



import ast
from importlib import import_module
from json import dumps, loads
from typing import TYPE_CHECKING, List, Optional, Type, Union

from nbformat import NotebookNode
from nbformat.v4 import new_code_cell, new_markdown_cell
from pydantic import BaseModel, SerializeAsAny, validator

if TYPE_CHECKING:
    from ..config import NBCXConfiguration
    from .context import NBCXContext


class NBCXType(BaseModel):
    module: str
    name: str

    @classmethod
    def from_string(cls, str: str) -> "NBCXType":
        module, name = str.split(":")
        return NBCXType(module=module, name=name)

    def to_string(self) -> str:
        return f"{self.module}:{self.name}"

    def type(self) -> Type["NBCXType"]:
        return getattr(import_module(self.module), self.name)

    def load(self, **kwargs) -> "NBCXType":
        return self.type()(**kwargs)


class NBCXBaseModel(BaseModel):
    type: SerializeAsAny[NBCXType]

    # internals
    _nb_var_name: Optional[str] = ""

    class Config:
        arbitrary_types_allowed: bool = False
        extra: str = "ignore"
        validate_assignment: bool = True

    def __init__(self, **kwargs):
        if "type" not in kwargs:
            kwargs["type"] = NBCXType(module=self.__class__.__module__, name=self.__class__.__name__)
        super().__init__(**kwargs)

    @validator("type", pre=True)
    def convert_type_string_to_module_and_name(cls, v):
        if isinstance(v, str):
            return NBCXType.from_string(v)
        return v

    @property
    def nb_var_name(self):
        if self._nb_var_name:
            return self._nb_var_name
        name = self.__class__.__name__.lower()
        if name.startswith("nbcx"):
            return name.replace("nbcx", "nbcx_")
        return f"nbcx_{name}"

    @staticmethod
    def _to_type(value, model_type=None):
        if value is None:
            value = {}

        if model_type is None and "type" in value:
            # derive type from instantiation
            model_type = NBCXBaseModel

        if isinstance(value, dict):
            return model_type(**value).type.load(**value)

        return value

    @classmethod
    def from_json(cls, json):
        data = loads(json)
        return cls(**data)

    def __call__(self, ctx: "NBCXContext" = None, *args, **kwargs) -> None:
        """Execute this model inside of a notebook

        Args:
            ctx (_type_, optional): _description_. Defaults to None.
        """
        ...

    def generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["NBCXConfiguration"] = None,
        parent: Optional["NBCXBaseModel"] = None,
        *args,
        **kwargs,
    ) -> Optional[Union[NotebookNode, List[NotebookNode]]]:
        """Generate a notebook node for this model.
        This will be called before the runtime of the notebook, use it for code generation.

        Args:
            metadata (dict): common cell metadata

        Returns:
            NotebookNode: the content of the notebook node
        """
        ...

    def _base_generate_meta(self, metadata: dict = None) -> Optional[NotebookNode]:
        cell = new_code_cell(metadata=metadata)
        cell.metadata.tags.extend(["nbcx"])
        cell.metadata.nbcx.role = "undefined"
        cell.metadata.nbcx.type = self.type.to_string()
        cell.metadata.nbcx.data = self.json()
        return cell

    def _base_generate_md_meta(self, metadata: dict = None) -> Optional[NotebookNode]:
        cell = new_markdown_cell(metadata=metadata)
        cell.metadata.tags.extend(["nbcx"])
        cell.metadata.nbcx.role = "undefined"
        cell.metadata.nbcx.type = self.type.to_string()
        return cell

    def _base_generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["NBCXConfiguration"] = None,
        parent: Optional["NBCXBaseModel"] = None,
        attr: str = "",
        counter: Optional[int] = None,
        call_with_context: str = "",
    ) -> Optional[NotebookNode]:
        cell = self._base_generate_meta(metadata=metadata)
        mod = ast.Module(body=[], type_ignores=[])

        if parent:
            value = ast.Attribute(value=ast.Name(id=parent.nb_var_name, ctx=ast.Load()), attr=attr, ctx=ast.Load())

            if counter is not None:
                value = ast.Subscript(
                    value=value,
                    slice=ast.Constant(value=counter),
                )
            mod.body.append(
                ast.Assign(
                    targets=[ast.Name(id=self.nb_var_name, ctx=ast.Store())],
                    value=value,
                    lineno=1,
                )
            )
        else:
            data = self.json()
            mod.body.append(
                ast.ImportFrom(
                    module=self.type.module,
                    names=[ast.alias(name=self.type.name)],
                    level=0,
                )
            )

            mod.body.append(
                ast.Assign(
                    targets=[ast.Name(id=self.nb_var_name, ctx=ast.Store())],
                    value=ast.Call(
                        func=ast.Attribute(
                            value=ast.Name(id=self.type.name, ctx=ast.Load()), attr="from_json", ctx=ast.Load()
                        ),
                        args=[ast.parse(dumps(data), mode="eval").body],
                        keywords=[],
                    ),
                    lineno=2,
                )
            )

        if call_with_context:
            mod.body.append(
                ast.Expr(
                    value=ast.Call(
                        func=ast.Name(id=self.nb_var_name, ctx=ast.Load()),
                        args=[],
                        keywords=[ast.keyword(arg="ctx", value=ast.Name(id=call_with_context, ctx=ast.Load()))],
                    )
                )
            )
        else:
            mod.body.append(ast.Expr(value=ast.Name(id=self.nb_var_name, ctx=ast.Load())))

        source = ast.unparse(mod).replace('"', '\\"')
        cell.source = source
        return cell

    def _base_generate_md(
        self,
        metadata: Optional[dict] = None,
    ) -> Optional[NotebookNode]:
        cell = self._base_generate_md_meta(metadata=metadata)
        return cell

    def __repr__(self) -> str:
        # Truncate the output for now
        return f"<{self.__class__.__name__}>"



from pathlib import Path
from typing import TYPE_CHECKING, Optional

from IPython.display import Image, display
from nbformat import NotebookNode
from pydantic import FilePath, validator

from ..base import NBCXBaseModel
from .base import NBCXContent

if TYPE_CHECKING:
    from ..config import NBCXConfiguration


class NBCXContentImage(NBCXContent):
    path: Optional[FilePath] = None
    content: Optional[bytes] = b""

    @validator("path", pre=True)
    def convert_path_from_obj(cls, v):
        if isinstance(v, str):
            v = Path(v).resolve()
        return v

    @validator("content", pre=True)
    def convert_content_from_obj(cls, v):
        if v is None:
            return b""
        if v and isinstance(v, str):
            v = Path(v).resolve()
        if v and isinstance(v, Path):
            v = v.read_bytes()
        return v

    def generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["NBCXConfiguration"] = None,
        parent: Optional[NBCXBaseModel] = None,
        attr: str = "",
        counter: Optional[int] = None,
    ) -> NotebookNode:
        # force parent to None to load from json explicitly
        cell = super()._base_generate(
            metadata=metadata,
            config=config,
            parent=parent,
            attr=attr,
            counter=counter,
            call_with_context=config.context.nb_var_name,
        )

        # add tags and role
        cell.metadata.tags.extend(["nbcx:content", "nbcx:content:dynamic"])
        cell.metadata.nbcx.role = "content"
        return cell

    def __call__(self, ctx=None, *args, **kwargs):
        if self.path:
            display(Image(filename=self.path))
        else:
            display(Image(data=self.content))

    def __repr__(self) -> str:
        return f"NBCXImage(path='{self.path}', content=[{len(self.content)} bytes])"



from typing import TYPE_CHECKING, Optional

from nbformat import NotebookNode
from pydantic import PrivateAttr

from .base import NBCXBaseModel

if TYPE_CHECKING:
    from .config import NBCXConfiguration


class NBCXContext(NBCXBaseModel):
    # internals
    _nb_var_name: str = PrivateAttr(default="nbcx_ctx")

    class Config:
        arbitrary_types_allowed: bool = True
        extra: str = "allow"
        validate_assignment: bool = False

    def generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["NBCXConfiguration"] = None,
        parent: Optional["NBCXBaseModel"] = None,
    ) -> NotebookNode:
        cell = self._base_generate(metadata=metadata)
        cell.metadata.tags.append("nbcx:context")
        cell.metadata.nbcx.role = "context"
        cell.metadata.nbcx.ignore = True
        return cell



from .base import NBCXContent, NBCXContentCode, NBCXContentDynamic, NBCXContentMarkdown
from .cover import NBCXContentCover
from .image import NBCXContentImage
from .pagebreak import NBCXContentPageBreak
from .table_of_contents import NBCXContentTableOfContents



class TestParity:
    def test_parity(self):
        import nbcx.utils.html as html
        import nbcx.utils.latex as latex

        _OMIT = (
            "Box",
            "BytesIO",
            "convert_pandoc",
            "display",
            "dt",
            "du",
            "html",
            "HTML",
            "Image",
            "latex",
            "Latex",
            "Output",
            "plot",
        )

        html_exps = [x for x in dir(html) if not (x in _OMIT or x.startswith("_"))]
        latex_exps = [x for x in dir(latex) if not (x in _OMIT or x.startswith("_"))]

        print(html_exps)
        print(latex_exps)
        assert set(html_exps) == set(latex_exps)



from nbconvert.exporters import LatexExporter
from nbconvert.filters.highlight import Highlight2Latex

from .template import TemplateOverrideMixin


class NBCXLatexExporter(TemplateOverrideMixin, LatexExporter):
    export_from_notebook = "NBCX LaTeX"

    def from_notebook_node(self, nb, resources=None, **kw):
        # ********************************************** #
        # From LatexExporter
        # https://github.com/jupyter/nbconvert/blob/master/nbconvert/exporters/latex.py
        langinfo = nb.metadata.get("language_info", {})
        lexer = langinfo.get("pygments_lexer", langinfo.get("name", None))
        highlight_code = self.filters.get("highlight_code", Highlight2Latex(pygments_lexer=lexer, parent=self))
        self.register_filter("highlight_code", highlight_code)
        # ********************************************** #

        # ***************** CUSTOM CODE **************** #
        return super()._from_notebook_node_override(nb, resources, **kw)
        # ********************************************** #



from typing import Dict, List, Optional

from pydantic import Field

from nbcx.config import NBCXParameters


class ExampleParameters(NBCXParameters):
    ticker: str
    a: Optional[int] = 4
    b: Optional[float] = 4.5
    c: List[int] = Field(default=[1, 2, 3])
    d: Dict[str, int] = Field(default={"A": 1, "B": 2, "C": 3})



import os
import os.path
import time

import pandas as pd
from IPython.display import HTML

from . import nbconvert_context
from .image import image

try:
    from tempfile import TemporaryDirectory
except BaseException:
    from backports.tempfile import TemporaryDirectory


with open(os.path.join(os.path.dirname(__file__), "resources", "jupyter_style.css"), "r") as fp:
    _JUPYTERLAB_STYLE = "<style>{}</style>".format(fp.read())


def _setup_screenshot(driver, path):
    """Grab screenshot of browser rendered HTML.
    Ensure the browser is sized to display all the HTML content."""
    # Ref: https://stackoverflow.com/a/52572919/
    driver.set_window_size(3000, 800)
    # required_width = driver.execute_script('return document.body.parentNode.scrollWidth')
    # required_height = driver.execute_script('return document.body.parentNode.scrollHeight')
    # driver.set_window_size(required_width, required_height)
    driver.save_screenshot(path)  # has scrollbar
    driver.find_element_by_tag_name("table").screenshot(path)  # avoids scrollbar
    # driver.find_element_by_tag_name('table').screenshot(path)  # avoids scrollbar
    # driver.set_window_size(original_size['width'], original_size['height'])


def _getTableImage(url, filename="dummy_table", path=".", delay=3, height=420, width=800):
    """Render HTML file in browser and grab a screenshot."""
    from selenium import webdriver
    from selenium.webdriver.chrome.options import Options

    options = Options()
    options.add_argument("--headless")
    options.add_argument("--hide-scrollbars")

    browser = webdriver.Chrome(chrome_options=options)
    browser.get(url)
    # Give the html some time to load
    time.sleep(delay)
    imgpath = "{}/{}.png".format(path, filename)
    _setup_screenshot(browser, imgpath)
    browser.quit()
    os.remove(imgpath.replace(".png", ".html"))
    return imgpath


def table_to_png(df_or_series_or_styler, **kwargs):
    if isinstance(df_or_series_or_styler, pd.DataFrame):
        tablehtml = df_or_series_or_styler.to_html()
    elif isinstance(df_or_series_or_styler, pd.Series):
        tablehtml = df_or_series_or_styler.to_html()
    else:
        tablehtml = df_or_series_or_styler.render()

    if nbconvert_context() == "pdf":
        with TemporaryDirectory() as dir:
            file_full_path = "{}/{}.html".format(dir, "temp")
            tmpurl = "file://{}".format(file_full_path)

            doc = '<html><head>{}</head><body><div class="jp-RenderedHTMLCommon">{}</div>'.format(
                _JUPYTERLAB_STYLE, tablehtml
            )
            with open(file_full_path, "w") as out:
                out.write(doc)
            with open(_getTableImage(tmpurl, "temp", dir), "rb") as fp:
                return image(fp.read())
    else:
        # defer to native rendering
        return HTML(tablehtml)



from .common import in_nbconvert, nbconvert_context  # noqa: F401

if nbconvert_context() == "pdf":
    from .latex import *  # noqa: F401, F403
else:
    from .html import *  # noqa: F401, F403

from .image import *  # noqa: F401, F403
from .pandas import table_to_png  # noqa: F401
